import collections

class AsaList(object):

    @classmethod
    def flatten(cls, lst):
        """
            Returns Generator of non-iterable values
        """
        for x in lst:
            if not isinstance(x, collections.Iterable):
                yield x
            else:
                for x in AsaList.flatten(x):
                    yield x



from paegan.cdm.dsg.collections.base.nested_point_collection import NestedPointCollection
from shapely.geometry import MultiPoint
from shapely.geometry import Point

class ProfileCollection(NestedPointCollection):
    """
        A collection of Profiles
    """

    def __init__(self, **kwargs):
        super(ProfileCollection,self).__init__(**kwargs)


from paegan.cdm.dsg.collections.base.feature_collection import FeatureCollection
from shapely.geometry import MultiPoint
from datetime import datetime
import sys

class PointCollection(FeatureCollection):
    """
        A collection of Points
    """

    def __init__(self, **kwargs):
        super(PointCollection,self).__init__(**kwargs)

    def filter_by_time(self, starting, ending):
        if not isinstance(starting, datetime) or not isinstance(ending, datetime):
            raise ValueError("Starting and ending must both be datetime objects")
        return (e for e in self.elements if starting <= e.time <= ending)

    def filter_by_variable(self, variable, elements=None):
        if elements is None:
            elements = self.elements

        return (m for m in elements for h in m.members if h.get("name","") == variable or h.get("standard","") == variable)

    def calculate_bounds(self):
        """
            Calculate the time_range, bbox, and size of this collection.
            Will scan all data.
            Ensures that .size, .bbox and .time_range return non-null.

            If the collection already knows its bbox; time_range; and/or size,
            they are recomputed.
        """ 
        stuff = map(lambda x: [x.time, x.location], self._elements)
        self.time_range = sorted(map(lambda x: x[0], stuff))
        points = map(lambda x: x[1], stuff)

        try:
            filtered_stuff = self.__filter_depths(stuff)
            self.depth_range = sorted(map(lambda x: x[1].z, filtered_stuff))
        except:
            self.depth_range = None

        try:
            self.bbox = MultiPoint(points).envelope
        except:
            self.bbox = None
            
        self.size = len(self._elements)


    def __filter_depths(self, list_to_filter):
        retval = list()
        for item in list_to_filter:
            try:
                getz = item[1].z
                retval.append(item)
            except:
                pass # don't do anything

        return retval



from shapely.geometry import Point

class FeatureCollection(object):
    """
        A collection of Features
        All Features in a FeatureCollection must be of the same type
    """

    def __init__(self, **kwargs):
        self.elements = kwargs.pop('elements', [])

    def set_type(self, type):
        self._type = type
    def get_type(self):
        return self._type
    type = property(get_type, set_type)

    def get_time_range(self):
        """ 
            Requires a call to calculateBounds
        """        
        return self._time_range
    def set_time_range(self, time_range):
        self._time_range = time_range
    time_range = property(get_time_range, set_time_range)

    def get_depth_range(self):
        """ 
            Requires a call to calculateBounds
        """        
        return self._depth_range
    def set_depth_range(self, depth_range):
        self._depth_range = depth_range
    depth_range = property(get_depth_range, set_depth_range)

    def lower_left(self):
        return Point(self._bbox.bounds[0], self._bbox.bounds[1])
    def upper_right(self):
        return Point(self._bbox.bounds[2], self._bbox.bounds[3])
    def get_bbox(self):
        """ 
            Returns a Shapely Polygon object representing the bbox
            
            Requires a call to calculateBounds
        """
        return self._bbox
    def set_bbox(self, bbox):
        self._bbox = bbox
    bbox = property(get_bbox, set_bbox)

    def get_size(self):
        """ 
            Requires a call to calculateBounds
        """        
        return self._size
    def set_size(self, size):
        self._size = size
    size = property(get_size, set_size)

    def add_element(self, element):
        self._elements.append(element)
    def set_elements(self, elements):
        self._elements = elements
    def get_elements(self):
        return self._elements
    elements = property(get_elements, set_elements)

    def __iter__(self):
        return iter(self.elements)


import math
from paegan.external.greatcircle import GreatCircle
import numpy as np

class AsaGreatCircle(object):

    @classmethod
    def great_circle(self, **kwargs):
        """
            Named arguments:
            distance = distance to traveled
            azimuth = angle, in DECIMAL DEGREES of HEADING from NORTH
            start_point = Location4D object representing the starting point
            rmajor = radius of earth's major axis. default=6378137.0 (WGS84)
            rminor = radius of earth's minor axis. default=6356752.3142 (WGS84)

            Returns a dictionary with:
            'latitude' in decimal degrees
            'longitude' in decimal degrees
            'reverse_azimuth' in decimal degrees

        """

        distance = kwargs.pop('distance')
        azimuth = kwargs.pop('azimuth')
        starting = kwargs.pop('start_point')
        rmajor = kwargs.pop('rmajor', 6378137.0)
        rminor = kwargs.pop('rminor', 6356752.3142)
        f = (rmajor - rminor) / rmajor

        lat_result, lon_result, angle_result = GreatCircle.vinc_pt(f, rmajor, math.radians(starting.latitude), math.radians(starting.longitude), math.radians(azimuth), distance)

        return {'latitude': math.degrees(lat_result), 'longitude': math.degrees(lon_result), 'reverse_azimuth': math.degrees(angle_result)}

    @classmethod
    def great_distance(self, **kwargs):
        """
            Named arguments:
            start_point = Location4D obect representing start point
            end_point = Location4D obect representing end point
            rmajor = radius of earth's major axis. default=6378137.0 (WGS84)
            rminor = radius of earth's minor axis. default=6356752.3142 (WGS84)

            Returns a dictionaty with:
            'distance' in meters
            'azimuth' in decimal degrees
            'reverse_azimuth' in decimal degrees

        """

        start_point = kwargs.pop('start_point', None)
        end_point = kwargs.pop('end_point', None)
        if start_point == None and end_point == None:
            start_lat = kwargs.pop("start_lats")
            start_lon = kwargs.pop("start_lons")
            end_lat = kwargs.pop("end_lats")
            end_lon = kwargs.pop("end_lons")
        rmajor = kwargs.pop('rmajor', 6378137.0)
        rminor = kwargs.pop('rminor', 6356752.3142)
        f = (rmajor - rminor) / rmajor

        if start_point != None and end_point != None:
            distance, angle, reverse_angle = GreatCircle.vinc_dist(f, rmajor, math.radians(start_point.latitude), 
                                                                   math.radians(start_point.longitude), 
                                                                   math.radians(end_point.latitude), 
                                                                   math.radians(end_point.longitude))
        else:
            vector_dist = np.vectorize(GreatCircle.vinc_dist)
            distance, angle, reverse_angle = vector_dist(f, rmajor, np.radians(start_lat), np.radians(start_lon),
                                                         np.radians(end_lat), np.radians(end_lon))
        return {'distance': distance, 'azimuth': np.degrees(angle), 'reverse_azimuth': np.degrees(reverse_angle)}
        
        



from shapely.geometry import Point
from copy import deepcopy

class Location4D(object):
    """
        A point in space and time
    """

    def __init__(self, **kwargs):
        """ Mandatory named arguments:
            * location (Location4d Object)
            OR
            * point (Shapely Point Object)
            OR 
            * latitude (DD)
            * longitude (DD)

            Optional named arguments: 
            * depth (meters)
            * time (DateTime Object)
        """
        if "location" in kwargs:
            loc = kwargs.pop('location')
            self.latitude = deepcopy(loc.latitude)
            self.longitude = deepcopy(loc.longitude)
            self.depth = deepcopy(loc.depth)
            self.time = deepcopy(loc.time)
        elif "point" in kwargs:
            self.point = kwargs.pop('point')
            self.time = kwargs.pop('time', None) 
        elif "latitude" and "longitude" in kwargs:
            self.latitude = kwargs.pop('latitude')
            self.longitude = kwargs.pop('longitude')
            self.depth = kwargs.pop('depth', None)
            self.time = kwargs.pop('time', None) 
        else:
            raise TypeError( "must provide a location4d object, a point geometry object, or a latitude and longitude" )

        self._dirty = True

    def set_point(self, point):
        self._point = point
        self._latitude = point.y
        self._longitude = point.x
        self._depth = point.z
        self._dirty = False
    def get_point(self):
        if self._dirty:
            self.point = Point(self._longitude, self._latitude, self._depth)
        return self._point
    point = property(get_point, set_point)

    def set_latitude(self, lat):
        self._latitude = lat
        self._dirty = True
    def get_latitude(self):
        return self._latitude
    latitude = property(get_latitude, set_latitude)

    def set_longitude(self, lon):
        self._longitude = lon
        self._dirty = True
    def get_longitude(self):
        return self._longitude
    longitude = property(get_longitude, set_longitude)

    def set_depth(self, dep):
        """
            Depth below sea level, positive up
        """
        self._depth = dep
        self._dirty = True
    def get_depth(self):
        return self._depth
    depth = property(get_depth, set_depth)

    def set_time(self, time):
        self._time = time
    def get_time(self):
        return self._time
    time = property(get_time, set_time)

    def __str__(self):
        return  "\n*** Location4D *** " + \
                "\nlatitude: " + str(self.latitude) + \
                "\nlongitude: " + str(self.longitude) + \
                "\ndepth: " + str(self.depth) + \
                "\ntime: " + str(self.time)

    def logstring(self):
        return "Time: %s, Lat: %f, Lon: %f, Depth: %fm " % (self.time, self.latitude, self.longitude, self.depth)

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
            and self.__dict__ == other.__dict__)


from math import sqrt, atan2, degrees, radians

class AsaMath(object):

    @classmethod
    def speed_direction_from_u_v(cls, **kwargs):
        if "u" and "v" in kwargs:
            speed = cls.__speed_from_u_v(kwargs.get('u'), kwargs.get('v'))
            direction = cls.__direction_from_u_v(kwargs.get('u'), kwargs.get('v'), output=kwargs.get('output'))
            return { 'speed':speed, 'direction':direction }
        else:
            raise TypeError( "must pass in 'u' and 'v' values ")

    @classmethod
    def __speed_from_u_v(cls, u, v):
        return sqrt((u*u) + (v*v))

    @classmethod
    def __direction_from_u_v(cls, u, v, **kwargs):
        rads = atan2(v, u)
        if 'output' in kwargs:
            if kwargs.pop('output') == 'radians':
                return rads

        # if 'output' was not specified as 'radians', we return degrees
        return cls.normalize_angle(angle=degrees(rads))

    @classmethod
    def azimuth_to_math_angle(cls, **kwargs):
        return cls.normalize_angle(angle=90 - kwargs.get("azimuth"))

    @classmethod
    def math_angle_to_azimuth(cls, **kwargs):
        return cls.normalize_angle(angle=(360 - kwargs.get("angle")) + 90)

    @classmethod
    def normalize_angle(cls, **kwargs):
        return kwargs.get('angle') % 360

    @classmethod
    def is_number(cls, num):
        try:
            float(num) # for int, long and float
        except TypeError:
            try:
                complex(num) # for complex
            except TypeError:
                return False

        return True



from paegan.cdm.dsg.collections.base.profile_collection import ProfileCollection

class Section(ProfileCollection):
    """
        A collection of profiles along a trajectory
    """

    def __init__(self, **kwargs):
        super(Section,self).__init__(**kwargs)
        self._type = "Section"

    def get_path(self):
        """
            Returns the nominal times of the profiles and the Points
            the profile was taken at.
        """
        return map(lambda x: [x.location, x.time], self._elements)


try:
    # Python 2.7
    from logging import NullHandler
except ImportError:
    # Python < 2.7
    import logging
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


from paegan.cdm.dsg.collections.base.nested_point_collection import NestedPointCollection

class TrajectoryCollection(NestedPointCollection):
    """
        A collection of Trajectories
    """

    def __init__(self, **kwargs):
        super(TrajectoryCollection,self).__init__(**kwargs)
        


# encoding: utf-8

# Pulled from Octant
# http://code.google.com/p/octant/source/browse/trunk/octant/ocean/time.py
# Copyright (c) 2006 Robert Hetland & Richard Hofmeister

import numpy as np
import netCDF4, datetime
from dateutil.parser import parse
import bisect
import pytz

# Same basedate as matplotlib: http://matplotlib.org/api/dates_api.html#matplotlib.dates.num2date
timevar_units = 'days since 0001-01-01 00:00:00'

def date2num(python_datetime):
    return netCDF4.date2num(python_datetime, timevar_units, 'proleptic_gregorian')
    
def num2date(indatenum, inunits, tzinfo=None):
    return np.vectorize(lambda x: x.replace(tzinfo=tzinfo))(netCDF4.num2date(indatenum, inunits, 'proleptic_gregorian'))

class Timevar(np.ndarray):
    
    _unit2sec={}
    _unit2sec['seconds'] = 1.0
    _unit2sec['minutes'] = 60.0
    _unit2sec['hours'] = 3600.0
    _unit2sec['days'] = 3600.0*24.0
    _unit2sec['weeks'] = 3600.0*24.0*7.0
    _unit2sec['years'] = 3600.0*24.0*365.242198781 #ref to udunits

    _sec2unit={}
    _sec2unit['seconds'] = 1.0
    _sec2unit['minutes'] = 1.0/60.0
    _sec2unit['hours'] = 1.0/3600.0
    _sec2unit['days'] = 1.0/(24.0*3600.0)

    def __new__(self, ncfile, name='time', units=None, tzinfo=None, **kwargs):
        if type(ncfile) is str:
            ncfile = netCDF4.Dataset(ncfile)
        self._nc = ncfile
        
        if self._nc.variables[name].ndim > 1:
            _str_data = self._nc.variables[name][:,:]
            if units == None:
                units = timevar_units
            dates = [parse(_str_data[i, :].tostring()) for i in range(len(_str_data[:,0]))]
            data = netCDF4.date2num(dates, units)
        else:
            data = self._nc.variables[name][:]
            
        if units == None:
            try:
                self._units = self._nc.variables[name].units
            except StandardError:
                self._units = units
        else:
            self._units = units

        if tzinfo == None:
            self._tzinfo = pytz.utc
        else:
            self._tzinfo = tzinfo

        
        units_split=self._units.split(' ',2)
        assert len(units_split) == 3 and units_split[1] == 'since', \
            'units string improperly formatted\n' + self._units
        self.origin=parse(units_split[2])
        
        self._units = units_split[0].lower()
        
        # compatibility to CF convention v1.0/udunits names:
        if self._units in ['second','sec','secs','s']:
            self._units='seconds'
        if self._units in ['min','minute','mins']:
            self._units='minutes'
        if self._units in ['h','hs','hr','hrs','hour']:
            self._units='hours'
        if self._units in ['day','d','ds']:
            self._units='days'

        return data.view(self)
    
    def gettimestep(self):
        return self.seconds[1] - self.seconds[0]
    
    def nearest_index(self, dateo, select='nearest'):
        to = date2num(dateo)
        if select == 'nearest':
            try:
                return [np.where(abs(self.datenum-t) == np.nanmin(abs(self.datenum-t)))[0][0] for t in to]
            except TypeError:
                return [np.where(abs(self.datenum-to) == np.nanmin(abs(self.datenum-to)))[0][0]]
        elif select == 'before':
            try: 
                return np.asarray([bisect.bisect(self.datenum, t)-1 for t in to])
            except TypeError:
                return np.asarray([bisect.bisect(self.datenum, to)-1])
    
    def nearest(self, dateo):
        """
        find nearest model timestep,
        input and output are datetime objects
        """
        # one might choose the second value for 
        #if len(self.nearest_index(dateo)) == 1:
        #    res=self.jd[self.nearest_index(dateo)][0]
        #else:
        #    res=self.jd[self.nearest_index(dateo)][1]
        return self.dates[self.nearest_index(dateo)][0]
    
    def get_seconds(self):
        fac = self._unit2sec[self._units] * self._sec2unit['seconds']
        return self*fac
    
    def get_minutes(self):
        fac = self._unit2sec[self._units] * self._sec2unit['minutes']
        return self*fac

    def get_hours(self):
        fac = self._unit2sec[self._units] * self._sec2unit['hours']
        return self*fac
    
    def get_days(self):
        fac = self._unit2sec[self._units] * self._sec2unit['days']
        return np.asarray(self,dtype='float64')*fac

    def get_dates(self):
        return num2date(self, self._units + " since " + self.origin.strftime('%Y-%m-%dT%H:%M:%S'), tzinfo=self._tzinfo)
    
    def get_datenum(self):
        return date2num(self.dates)
        
    datenum = property(get_datenum, None, doc="datenum in seconds since 1970-01-01")
    seconds = property(get_seconds, None, doc="seconds")
    minutes = property(get_minutes, None, doc="minutes")
    hours = property(get_hours, None, doc="hours")
    days = property(get_days, None, doc="days")
    dates = property(get_dates, None, doc="datetime objects")
    timestep = property(gettimestep, None)



import logging
import collections
import multiprocessing
import threading

from paegan.logger.multi_process_logging import MultiProcessingLogHandler
from paegan.logger.progress_handler import ProgressHandler

class EasyLogger(object):
    def __init__(self, logpath, level=None):

        if level is None:
            level = logging.PROGRESS

        self.logpath = logpath
        self.logger = multiprocessing.get_logger()

        self.queue = multiprocessing.Queue(-1)

        # Close any existing handlers
        self.close_handlers()
        
        # Remove any existing handlers
        self.logger.handlers = []
        self.logger.setLevel(level)

        handler = MultiProcessingLogHandler(logpath, self.queue)
        handler.setLevel(level)
        formatter = logging.Formatter('[%(asctime)s] - %(levelname)s - %(processName)s - %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        progress_deque = collections.deque(maxlen=1)
        progress_handler = ProgressHandler(progress_deque)
        progress_handler.setLevel(level)
        self.logger.addHandler(progress_handler)
        self.e = threading.Event()

        def save_progress():
            while self.e.wait(5) != True:
                try:
                    record = progress_deque.pop()
                    if record == StopIteration:
                        break

                    print record

                except Exception:
                    pass
            return

        t = threading.Thread(name="ProgressUpdater", target=save_progress)
        t.daemon = True
        t.start()

    def close_handlers(self):
        (hand.close() for hand in self.logger.handlers)
        
    def close_queue(self):
        self.queue.put_nowait(StopIteration)
        self.e.set()

    def close(self):
        self.close_handlers()
        self.close_queue()

    def __str__(self):
        return "Logging with MultiProcessingLogHandler in " + self.logpath
        


import numpy as np
import netCDF4

class Depthvar(np.ndarray):

    # How many meter in the unit
    _unit2meters={}
    _unit2meters['millimeters'] = 0.001
    _unit2meters['centimeters'] = 0.01
    _unit2meters['meters'] = 1
    _unit2meters['feet'] = 0.3048
    _unit2meters['yards'] = 0.9144
    _unit2meters['kilometers'] = 1000
    _unit2meters['miles'] = 1609.34

    # How many units in the meter
    _meters2unit={}
    _meters2unit['millimeters'] = 1000
    _meters2unit['centimeters'] = 100
    _meters2unit['meters'] = 1
    _meters2unit['feet'] = 3.2084
    _meters2unit['yards'] = 1.09361
    _meters2unit['kilometers'] = 0.001
    _meters2unit['miles'] = 0.000621371

    def __new__(self, ncfile, name, units=None, **kwargs):
        if type(ncfile) is str:
            ncfile = netCDF4.Dataset(ncfile)
        self._nc = ncfile
        
        data = self._nc.variables[name][:]
        if units == None:
            try:
                self._units = self._nc.variables[name].units
            except StandardError:
                self._units = 'meters'
        else:
            self._units = units
               
        # compatibility to CF convention v1.0/udunits names:
        if self._units in ['m','meter','meters from the sea surface']:
            self._units='meters'
        if self._units in ['cm','centimeter']:
            self._units='centimeters'
        if self._units in ['mm','millimeter']:
            self._units='millimeters'
        if self._units in ['km','kilometer']:
            self._units='kilometers'
        if self._units in ['ft','feets']:
            self._units='feet'
        if self._units in ['yd','yard']:
            self._units='yards'
        if self._units in ['mile']:
            self._units='miles'

        return data.view(self)
    
    def nearest_index(self, depth):
        return np.where(abs(self.meters-depth) == np.nanmin(abs(self.meters-depth)))[0]
    
    def nearest(self, depth):
        """
        find nearest depth,
        input and output are meters
        """
        return self.meters[self.nearest_index(depth)][0]
    
    def get_mm(self):
        fac = self._unit2meters[self._units] * self._meters2unit['millimeters']
        return self*fac
    
    def get_cm(self):
        fac = self._unit2meters[self._units] * self._meters2unit['centimeters']
        return self*fac

    def get_km(self):
        fac = self._unit2meters[self._units] * self._meters2unit['kilometers']
        return self*fac

    def get_m(self):
        fac = self._unit2meters[self._units] * self._meters2unit['meters']
        return self*fac
        
    meters = property(get_m, None, doc="meters")
    kilometers = property(get_km, None, doc="kilometers")
    centimeters = property(get_cm, None, doc="centimeters")
    millimeters = property(get_mm, None, doc="millimeters")



import random
import uuid

class AsaRandom(object):

    @staticmethod
    def random():
        """
        pseudo random number generator
        """
        rand = random.random()
        flip = random.random()
        if flip > 0.5:
            rand *= -1
        return rand

    @staticmethod
    def filename(prefix=None, suffix=None):
        fn = []
        if prefix:
            fn.extend([prefix, '-'])

        fn.append(str(uuid.uuid4()))

        if suffix:
            fn.extend(['.', suffix.lstrip('.')])

        return ''.join(fn)



from paegan.cdm.dsg.collections.base.point_collection import PointCollection

class Trajectory(PointCollection):
    """
        A collection of Points along a 1 dimensional path, connected in space and time.
        The Points are ordered in time (in other words, the time dimension must
        increase monotonically along the trajectory).
    """

    def __init__(self, **kwargs):
        super(Trajectory,self).__init__(**kwargs)
        self._type = "Trajectory"

    def get_path(self):
        """
            Returns the Points along the trajectory
        """
        return map(lambda x: [x.location, x.time], self._elements)


import numpy as np

class Coordinates(object):
    """
    This is just a container to hold the coordinate variables persistently in the dataset obj for each of the field variables of interest.
    """
    def __init__(self, **kwargs):
        self.xy, self.z, self.time = None, None, None
    
        if "gridobj" in kwargs:
            self.xy = kwargs["xy"]
        if "depthvar" in kwargs:
            self.z = kwargs["z"]
        if "timevar" in kwargs:
            self.time = kwargs["time"]
        
        
            
    def add_xy(self, gridobj):
        self.xy = gridobj
        
    def add_z(self, depthvar):
        self.z = depthvar
        
    def add_time(self, timevar):
       self.time = timevar
    
    def _getinfo(self):
       info = ""
       if self.xy != None:
           info = info + "[XY]"
       if self.z != None:
           info = info + "[Z]"
       if self.time != None:
           info = info + "[T]"
       return info
       
    def get_xarray(self):
        return self.xy._xarray
        
    def get_yarray(self):
        return self.xy._yarray
        
    def __str__(self):
        return self._info
    
    def __unicode__(self):
        return self._info
        
    _info = property(_getinfo)
    x = property(get_xarray)
    y = property(get_yarray)
    
    
class SubCoordinates(object):
    def __init__(self, **kwargs):
        self.x = None
        self.y = None
        self.z = None
        self.time = None
        if "x" in kwargs:
            self.x = kwargs["x"]
        if "y" in kwargs:
            self.y = kwargs["y"]
        if "z" in kwargs:
            self.z = kwargs["z"]
        if "time" in kwargs:
            self.time = kwargs["time"]
       
       
         
            
            



import pytz
import time
import logging

from datetime import datetime
from collections import deque

class OnlyProgressFilter(logging.Filter):
    def filter(self, record):
        return record.levelno == logging.PROGRESS

class ProgressHandler(logging.Handler):
    def __init__(self, progress_deque):
        logging.Handler.__init__(self)
        self.addFilter(OnlyProgressFilter())
        self._progress_deque = progress_deque      

    def send(self, s):
        self._progress_deque.append(s)

    def emit(self, record):
        try:
            s = self._format_record(record)
            self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)

    def _format_record(self, record):
        # format log into tuple of:
        # (datetime, progress, message)

        # Get a timezone aware datetime from time.time()
        dt = datetime.fromtimestamp(record.created).replace(tzinfo=pytz.timezone(time.strftime("%Z",time.gmtime()))).astimezone(pytz.utc)

        msg = record.msg

        if isinstance(msg, list):
            msg = tuple(msg)

        if isinstance(msg, tuple):
            return tuple([dt]) + msg
        elif isinstance(msg, str) or isinstance(msg, unicode):
            return (dt, -1, unicode(msg))
        elif isinstance(msg, float) or isinstance(msg, int):
            return (dt, msg, None)
        else:
            return (dt, None, None)

    def close(self):
        self._progress_deque.append(StopIteration)
        logging.Handler.close(self)


from shapely.geometry import MultiPoint

from paegan.cdm.dsg.collections.base.point_collection import PointCollection

class Station(PointCollection):
    """
        A collection of points at a single location
    """
    def __init__(self, **kwargs):
        super(Station,self).__init__(**kwargs)
        self._type = "timeSeries"
        self._properties = dict()
        self.uid = None
        self.name = None        
        self.description = None

    def get_location(self):
        return self._location
    def set_location(self, location):
        self._location = location

        # Sets the location of every Point in this Station if it is not already set
        for p in self._elements:
            try:
                assert p.location is not None
            except:
                p.location = self._location
        
    location = property(get_location, set_location)

    def get_uid(self):
        return self._uid
    def set_uid(self, uid):
        self._uid = uid
    uid = property(get_uid, set_uid)

    def get_name(self):
        return self._name
    def set_name(self, name):
        self._name = name
    name = property(get_name, set_name)

    def get_unique_members(self):
        all_members = (m for m in (e.members for e in self.elements))

        keys = ["name", "description", "standard", "unit"]
        mwhat = []
        for mg in all_members:
            for m in mg:
                mwhat.append( { key: m[key] for key in keys if key in m } )

        # Now unique them on name
        mwhat = { x['name']:x for x in mwhat }.values()

        return mwhat

    def get_description(self):
        return self._description
    def set_description(self, description):
        self._description = description
    description = property(get_description, set_description)

    def properties(self):
        """ General properties to store things about a station """
        return self._properties
    def set_property(self, prop, value):
        self._properties[prop] = value
    def get_property(self, prop):
        return self._properties.get(prop, None)

    def calculate_bounds(self):
        """
            Calculate the time_range, bbox, and size of this collection.
            Will scan all data.
            Ensures that .size, .bbox and .time_range return non-null.

            If the collection already knows its bbox; time_range; and/or size,
            they are recomputed.
        """ 
        self.location = self._location # To set locations for all points that don't have one
        stuff = map(lambda x: [x.time, x.location], self._elements)
        self.time_range = sorted(map(lambda x: x[0], stuff))
        points = map(lambda x: x[1], stuff)
        try:
            self.depth_range = sorted(map(lambda x: x[1].z, stuff))
        except:
            self.depth_range = None
        self.bbox = MultiPoint([self.location, self.location]).envelope
        self.size = len(self._elements)



from paegan.cdm.dsg.collections.base.nested_point_collection import NestedPointCollection
from paegan.cdm.dsg.features.station import Station

class StationCollection(NestedPointCollection):
	""" Collection of Station features """
	def __init__(self, **kwargs):
		super(StationCollection,self).__init__(**kwargs)



import os.path
import threading
import numpy as np
import netCDF4

#       ---------------------------------
#       rho | u | rho | u | rho | u | rho
#       ---------------------------------
#        v  |   | {v} |   | {v} |   |  v
#       ---------------------------------
#       rho |{u}|(rho)|{u}|(rho)|{u}| rho
#       ---------------------------------
#        v  |   | {v} |   | {v} |   |  v
#       ---------------------------------
#       rho | u | rho | u | rho | u | rho
#       ---------------------------------
#       
#       Assumes:
#       size(rho)=ny,nx;   size(u)=ny,nx-1;   size(v)=ny-1,nx
#       coordinate dimension: t,z,y,x
#
#       When Averaging U and V into the RHO grid:
#       * We lose the first and last column of rho and V
#       * We lose the first and last row of rho and U

def shrink(a, b):
    """
    Ripped from Octant: https://code.google.com/p/octant

    Return array shrunk to fit a specified shape by triming or averaging.
    
    a = shrink(array, shape)
    
    array is an numpy ndarray, and shape is a tuple (e.g., from
    array.shape). a is the input array shrunk such that its maximum
    dimensions are given by shape. If shape has more dimensions than
    array, the last dimensions of shape are fit.
    
    as, bs = shrink(a, b)
    
    If the second argument is also an array, both a and b are shrunk to
    the dimensions of each other. The input arrays must have the same
    number of dimensions, and the resulting arrays will have the same
    shape.
    
    Example
    -------
    
    >>> shrink(rand(10, 10), (5, 9, 18)).shape
    (9, 10)
    >>> map(shape, shrink(rand(10, 10, 10), rand(5, 9, 18)))        
    [(5, 9, 10), (5, 9, 10)]   
       
    """

    if isinstance(b, np.ndarray):
        if not len(a.shape) == len(b.shape):
            raise Exception, \
                  'input arrays must have the same number of dimensions'
        a = shrink(a,b.shape)
        b = shrink(b,a.shape)
        return (a, b)

    if isinstance(b, int):
        b = (b,)
        
    if len(a.shape) == 1:                # 1D array is a special case
        dim = b[-1]                      
        while a.shape[0] > dim:          # only shrink a
            if (dim - a.shape[0]) >= 2:  # trim off edges evenly
                a = a[1:-1]
            else:                        # or average adjacent cells
                a = 0.5*(a[1:] + a[:-1])
    else:
        for dim_idx in range(-(len(a.shape)),0):
            dim = b[dim_idx]
            a = a.swapaxes(0,dim_idx)        # put working dim first
            while a.shape[0] > dim:          # only shrink a
                if (a.shape[0] - dim) >= 2:  # trim off edges evenly
                    a = a[1:-1,:]
                if (a.shape[0] - dim) == 1:  # or average adjacent cells
                    a = 0.5*(a[1:,:] + a[:-1,:])
            a = a.swapaxes(0,dim_idx)        # swap working dim back
        
    return a

def uv_to_rho(file):
    nc = netCDF4.Dataset(file)

    lat_rho=nc.variables['lat_rho']
    lon_rho=nc.variables['lon_rho']

    # lat_rho and lon_rho should have identical shapes (eta,xi)/(y/x)
    if lat_rho.shape != lon_rho.shape:
        print "Shape of lat_rho and lon_rho must be equal"
        return None

    # Store shape for use below
    [rho_y,rho_x] = lat_rho.shape

    # U
    u_data = nc.variables['u'][0,0,:,:]

    # V
    v_data = nc.variables['v'][0,0,:,:]

    # Get the angles
    angle = nc.variables['angle'][:,:]

    # Create empty rho matrix to store complex U + Vj
    U = np.empty([rho_y,rho_x], dtype=complex )
    # And fill it with with nan values
    U[:] = np.nan

    vfunc = np.vectorize(complex)

    # THREE IDENTICAL METHODS
    # Fill in RHO cells per diagram above.  Skip first row and first
    # column of RHOs and leave them as numpy.nan values.  Also skip the 
    # last row and column.

    #1.) 
    # Fill rho points with averages (if we can do the calculation)
    # Thread]
    t1 = AverageAdjacents(u_data)
    t2 = AverageAdjacents(v_data, True)
    t1.start(); t2.start()
    t1.join();  t2.join()
    u_avg = t1.data
    v_avg = t2.data
    complexed = vfunc(u_avg[1:-1,:],v_avg[:,1:-1])

    # 2.) 
    # Don't thread
    #u_avg = average_adjacents(u_data)
    #v_avg = average_adjacents(v_data,True)
    #complexed = vfunc(u_avg[1:-1,:],v_avg[:,1:-1])

    # 3.)
    # Use Octant method
    #data_U = (rho_y-2,rho_x-2)
    #u_avg = shrink(u_data, data_U)
    #v_avg = shrink(v_data, data_U)
    #complexed = vfunc(u_avg,v_avg)
    
    # Only pull the U and V values that can contribute to the averaging (see diagram).
    # This means we lost the first and last row and the first and last column of both
    # U and V.   
    U[1:rho_y-1, 1:rho_x-1] = complexed
 
    # We need the rotated point, so rotate by the "angle"
    return rotate_complex_by_angle(U,angle)

def rotate_complex_by_angle(points,angles):
    """
        points: ND array of complex numbers
        angles: ND array of angles, same size as points

        Multiply the points we have (stored as complex) by another complex number
        whose angle with the positive X axis is the angle we need to rotate by.

        np.exp(1J*angle)  creates an array the same size of U filled with complex numbers
                          that represent the angled points from X axis.                    
    """
    return points * np.exp(1j*angles)

def average_adjacents(a, by_column=False):
    """
        Sums adjacent values in a column.  Optional by_column parameter
        will sum adjacement column values.

        For a row that looks like this:
        [ 2, 4, 6, 8, 10, 12 ]

        The following addition is computed:

        [ 2, 4, 6,  8, 10 ]
                +
        [ 4, 6, 8, 10, 12 ]

        Resulting in

        [ 3, 5, 7, 9, 11 ]

        Use:

            # 1D Array
            >>> a = np.arange(0,13,2)
            >>> a
            array([ 0,  2,  4,  6,  8, 10, 12])

            >>> a_avg = average_adjacents(a)
            >>> a_avg
            array([  1.,   3.,   5.,   7.,   9.,  11.])
            

            # 2D Array
            >>> a = np.arange(0,30,2).reshape(3,5)
            >>> a
            array([[ 0,  2,  4,  6,  8],
                   [10, 12, 14, 16, 18],
                   [20, 22, 24, 26, 28]])
            >>> a_avg = average_adjacents(a)
            >>> a_avg
            array([[  1,  3,  5,  7],
                   [ 11, 13, 15, 17],
                   [ 21, 23, 25, 27]])  


            # 2D Array, by column
            >>> a = np.arange(0,30,2).reshape(3,5)
            >>> a
            array([[ 0,  2,  4,  6,  8],
                   [10, 12, 14, 16, 18],
                   [20, 22, 24, 26, 28]])
            >>> a_avg = average_adjacents(a, True)
            >>> a_avg
            array([[  5.,   7.,   9.,  11.,  13.],
                   [ 15.,  17.,  19.,  21.,  23.]])

    """
    if by_column:
        # Easier to transpose, sum by row, and transpose back.
        #sumd = 0.5 * (a[1:m,:] + a[0:m-1,:])
        return average_adjacents(a.T).T

    try:
        [m,n] = a.shape
    except ValueError:
        [m,n] = [1,a.size]

    if m == 1:
        sumd = 0.5 * (a[0:n-1] + a[1:n]) # Single row
    else:
        sumd = 0.5 * (a[:,0:n-1] + a[:,1:n]) # By row

    return sumd

# Threaded instance for speed testing on enormous grids.
class AverageAdjacents(threading.Thread):
    def __init__(self, data, by_column=False):
        threading.Thread.__init__(self)
        self.by_column = by_column
        self.data = data
    def run(self):
        if self.by_column:
            # Easier to transpose, sum by row, and transpose back.
            #sumd = 0.5 * (a[1:m,:] + a[0:m-1,:])
            t3 = AverageAdjacents(self.data.T)
            t3.start(); t3.join()
            self.data = t3.data.T
            return

        try:
            [m,n] = self.data.shape
        except ValueError:
            [m,n] = [1,self.data.size]

        if m == 1:
            sumd = 0.5 * (self.data[0:n-1] + self.data[1:n]) # Single row
        else:
            sumd = 0.5 * (self.data[:,0:n-1] + self.data[:,1:n]) # By row

        self.data = sumd
        return


from logging import FileHandler
import multiprocessing, threading, logging, sys, traceback

class MultiProcessingLogHandler(logging.Handler):
    def __init__(self, name, queue):
        logging.Handler.__init__(self)

        self._handler = FileHandler(name)
        self.queue = queue

        t = threading.Thread(target=self.receive)
        t.daemon = True
        t.start()

    def setFormatter(self, fmt):
        logging.Handler.setFormatter(self, fmt)
        self._handler.setFormatter(fmt)

    def receive(self):
        while True:
            try:
                record = self.queue.get()
                if record == StopIteration:
                    break
                self._handler.emit(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except Exception:
                traceback.print_exc(file=sys.stderr)
                break

        return

    def send(self, s):
        self.queue.put_nowait(s)

    def _format_record(self, record):
        # ensure that exc_info and args
        # have been stringified.  Removes any chance of
        # unpickleable things inside and possibly reduces
        # message size sent over the pipe
        if record.args:
            record.msg = record.msg % record.args
            record.args = None
        if record.exc_info:
            dummy = self.format(record)
            record.exc_info = None

        return record

    def emit(self, record):
        try:
            s = self._format_record(record)
            self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)

    def close(self):
        self.queue.put_nowait(StopIteration)
        self._handler.close()
        logging.Handler.close(self)



import numpy as np
import netCDF4, datetime, copy
from paegan.cdm.timevar import Timevar
from paegan.cdm.depthvar import Depthvar
from paegan.cdm.gridvar import Gridobj
from paegan.cdm.variable import Coordinates as cachevar
from paegan.cdm.variable import SubCoordinates as subs
from paegan.location4d import Location4D

from paegan.logger import logger

_possiblet = ["time", "TIME", "Time",
           "t", "T",
           "ocean_time", "OCEAN_TIME",
           "jd", "JD",
           "dn", "DN",
           "times", "TIMES", "Times",
           "mt", "MT",
           "dt", "DT",
          ]
_possiblez = ["depth", "DEPTH",
           "depths", "DEPTHS",
           "height", "HEIGHT",
           "altitude", "ALTITUDE",
           "alt", "ALT", 
           "Alt", "Altitude",
           "h", "H",
           "s_rho", "S_RHO",
           "s_w", "S_W",
           "z", "Z",
           "siglay", "SIGLAY",
           "siglev", "SIGLEV",
           "sigma", "SIGMA",
          ]
_possiblex = ["x", "X",
           "lon", "LON",
           "xlon", "XLON",
           "lonx", "lonx",
           "lon_u", "LON_U",
           "lon_v", "LON_V",
           "lonc", "LONC",
           "Lon", "Longitude",
           "longitude", "LONGITUDE",
           "lon_rho", "LON_RHO",
           "lon_psi", "LON_PSI",
          ]
_possibley = ["y", "Y",
           "lat", "LAT",
           "ylat", "YLAT",
           "laty", "laty",
           "lat_u", "LAT_U",
           "lat_v", "LAT_V",
           "latc", "LATC",
           "Lat", "Latitude",
           "latitude", "LATITUDE",
           "lat_rho", "LAT_RHO",
           "lat_psi", "LAT_PSI",
          ]

def _sub_by_nan(data, ind):
        """
            Funtction to subset a dimension variable by replacing values
            that do not appear in the index with np.nan, in order to
            preserve the lazy data access on the full array's in the backend.
        """
        if len(ind) > 0:
            xtmp = -1 * np.ones_like(data)
            xtmp[ind[0]:ind[-1]+1] = ind
            xbool = range(len(data)) != xtmp
            data[xbool] = np.nan
        else:
            data = np.nan * np.ones_like(data)
        return data
        
def _sub_by_nan2(data, ind):
        """
            Funtction to subset a dimension variable by replacing values
            that do not appear in the index with np.nan, in order to
            preserve the lazy data access on the full array's in the backend.
        """
        if (len(ind[0]) > 0) & (len(ind[1]) > 0):
            xtmp = -1 * np.ones_like(data[1,:])
            xtmp[ind[1][0]:ind[1][-1]+1] = ind[1]
            xbool = range(len(data[1])) != xtmp
            ytmp = -1 * np.ones_like(data[:,1])
            ytmp[ind[0][0]:ind[0][-1]+1] = ind[0]
            ybool = range(len(data[0])) != ytmp
            data[ybool, xbool] = np.nan
        else:
            data = np.nan * np.ones_like(data)
        return data
        
class CommonDataset(object):

    @staticmethod
    def open(ncfile, xname='lon', yname='lat', zname='z', tname='time', **kwargs):
        """
        Initialize paegan dataset object, which uses specific
        readers for different kinds of datasets, and returns
        dataset objects that expose a common api.
        
        from cdm.dataset import CommonDataset

        >> dataset = CommonDataset.open(ncfile)
        >> dataset = CommonDataset.open(url, "lon_rho", "lat_rho", "s_rho", "ocean_time")
        >> dataset = CommonDataset.open(url, dataset_type="cgrid")
        """

        nc = None
        filename = None

        if isinstance(ncfile, str):
            ncfile = unicode(ncfile.strip())

        if isinstance(ncfile, unicode):
            try:
                nc = netCDF4.Dataset(ncfile)
                filename = ncfile
            except StandardError:
                logger.error(ncfile)
                raise
        elif isinstance(ncfile, Dataset):
            # Passed in paegan Dataset object
            nc = ncfile.nc
        elif isinstance(ncile, netCDF4.Dataset):
            # Passed in a netCDF4 Dataset object
            nc = ncfile

        datasettype = kwargs.get('dataset_type', None)
        
        # Find the coordinate variables for testing, unknown if not found
        keys = set(nc.variables)
        posx = set(_possiblex)
        posy = set(_possibley)
        xmatches = list(posx.intersection(keys))
        ymatches = list(posy.intersection(keys))
        
        if xname in keys and yname in keys:
            testvary = nc.variables[yname]
            testvarx = nc.variables[xname]
        elif len(xmatches) > 0:
            testvary = nc.variables[ymatches[0]]
            testvarx = nc.variables[xmatches[0]]
        
        # Test the shapes of the coordinate variables to determine the grid type
        if datasettype is None:
            if testvary.ndim > 1:
                datasettype = "cgrid"
            else:
                if testvary.shape[0] != testvarx.shape[0]:
                    datasettype = "rgrid"
                else:
                    if "cdm_data_type" in nc.ncattrs():
                        if nc.cdm_data_type.lower() == "grid":
                            datasettype = "rgrid"
                        else:
                            datasettype = "ncell"
                    else:
                        datasettype = "ncell"
        nc.close()

        # Return appropriate dataset subclass based on datasettype
        from paegan.cdm.grids.c_grid import CGridDataset
        from paegan.cdm.grids.n_cell import NCellDataset
        from paegan.cdm.grids.u_grid import UGridDataset
        from paegan.cdm.grids.r_grid import RGridDataset

        if datasettype == 'ncell':
            dataobj = NCellDataset(filename, datasettype,
                zname=zname, tname=tname, xname=xname, yname=yname)
        elif datasettype == 'rgrid':
            dataobj = RGridDataset(filename, datasettype,
                zname=zname, tname=tname, xname=xname, yname=yname)
        elif datasettype == 'cgrid':
            dataobj = CGridDataset(filename, datasettype,
                zname=zname, tname=tname, xname=xname, yname=yname)
        else:
            dataobj = None

        return dataobj
    

class Dataset(object):
    def __init__(self, filename, datasettype, xname='lon', yname='lat',
        zname='z', tname='time'):
        self.nc = None
        self._coordcache = dict()
        self._filename = filename
        self._datasettype = datasettype
        
        self._possiblet = _possiblet
        self._possiblez = _possiblez
        self._possiblex = _possiblex
        self._possibley = _possibley
                
        if xname not in self._possiblex:
            self._possiblex.append(xname)
        if yname not in self._possibley:
            self._possibley.append(yname)
        if zname not in self._possiblez:
            self._possiblez.append(zname)
        if tname not in self._possiblet:
            self._possiblet.append(tname)

        self.opennc()
        self._current_variables = list(self.nc.variables.keys())
        
    def _copy(self):
        raise NotImplementedError
        
    """
    
        Methods that return data or info or something
        
    """
    def getvariableinfo(self):
        variables = {}
        for var in self._current_variables:
            variables[var] = {}
            for attr in self.nc.variables[var].ncattrs():
                variables[var][attr] = self.nc.variables[var].getncattr(attr)
        return variables
                           
    def lon2ind(self, var=None, **kwargs):
        raise NotImplementedError
         
    def lat2ind(self, var=None, **kwargs):
        raise NotImplementedError
            
    def ind2lon(self, var=None, **kwargs):
        raise NotImplementedError
   
    def ind2lat(self, var=None, **kwargs):
        raise NotImplementedError
    
    def get_xyind_from_bbox(self, var, bbox):
        raise NotImplementedError
        
    def get_xyind_from_point(self, var, point, **kwargs):
        raise NotImplementedError
        
    def closenc(self):
        self.metadata = None
        self.nc.close()
        self.nc = None
        
    def opennc(self):
        self.nc = netCDF4.Dataset(self._filename)
        self.metadata = self.nc.__dict__
    
    def gettimestep(self, var=None):
        assert var in self._current_variables
        time = self.gettimevar(var)
        return time[np.isnan(time)==False].timestep
    
    def gettimebounds(self, var=None, **kwargs):
        assert var in self._current_variables
        time = self.gettimevar(var)
        if "units" in kwargs:
            u = kwargs.get("units")
            bounds = (netCDF4.num2date(np.nanmin(time[np.isnan(time)==False]),units=u),
                      netCDF4.num2date(np.nanmax(time[np.isnan(time)==False]),units=u))
        else:
            bounds = (np.nanmin(time[np.isnan(time)==False].dates), np.nanmax(time[np.isnan(time)==False].dates))
        return bounds

    def getdepthbounds(self, var=None, **kwargs):
        assert var in self._current_variables
        depths = self.getdepthvar(var)
        if "units" in kwargs:
            if kwargs["units"] == "m":
                bounds = (np.nanmin(depths.get_m), np.nanmax(depths.get_m))
            else:
                bounds = ()
        else:
            bounds = (np.nanmin(depths), np.nanmax(depths))
        return bounds
    
    def getbbox(self, var=None, **kwargs):
        assert var in self._current_variables
        grid = self.getgridobj(var)
        return grid.bbox
        
    def getboundingpolygon(self, var=None, **kwargs):
        assert var in self._current_variables
        grid = self.getgridobj(var)
        return grid.boundingpolygon
        
    def _checkcache(self, var):
        assert var in self._current_variables
        test = var in self._coordcache
        return test

    def gettimevar(self, var=None, use_cache=True):
        #return self._timevar
        assert var in self._current_variables
        timevar = None
        if use_cache == True:
            if self._checkcache(var):
                timevar = self._coordcache[var].time
            else:
                self._coordcache[var] = cachevar()
        if timevar == None:
            names = self.get_coord_names(var)
            if names['tname'] != None:
                timevar = Timevar(self.nc, names["tname"])
            else:
                timevar = None
            if use_cache == True:  
                self._coordcache[var].time = timevar
        return timevar
 
    def getdepthvar(self, var=None, use_cache=True):
        #return self._depthvar
        assert var in self._current_variables
        depthvar = None
        if use_cache == True:
            if self._checkcache(var):
                depthvar = self._coordcache[var].z
            else:
                self._coordcache[var] = cachevar()
        if depthvar == None:
            names = self.get_coord_names(var)
            if names['zname'] != None:
                depthvar = Depthvar(self.nc, names["zname"])
            else:
                depthvar = None
            if use_cache == True:
                self._coordcache[var].add_z(depthvar)
        return depthvar
        
    def getgridobj(self, var=None):
        #return self._gridobj
        assert var in self._current_variables
        gridobj = None
        if self._checkcache(var):
            gridobj = self._coordcache[var].xy
        else:
            self._coordcache[var] = cachevar()
        if gridobj == None:
            names = self.get_coord_names(var)
            if names['xname'] != None and names['yname'] !=None:
                gridobj = Gridobj(self.nc, names["xname"], 
                   names["yname"])
            else:
                gridobj = None
            self._coordcache[var].add_xy(gridobj)
        return gridobj
        
    def get_tind_from_bounds(self, var, bounds, convert=False, use_cache=True):
        assert var in self._current_variables
        time = self.gettimevar(var, use_cache)
        if convert:
            bounds = netCDF4.date2num(bounds, time._units + " since " + time.origin.isoformat())
        inds = np.where(np.logical_and(time.dates >= bounds[0].replace(tzinfo=time.dates[0].tzinfo), time.dates <= bounds[1].replace(tzinfo=time.dates[0].tzinfo)))
        return inds
    
    def get_zind_from_bounds(self, var, bounds, use_cache=True):
        assert var in self._current_variables
        depths = self.getdepthvar(var, use_cache)
        inds = np.where(np.logical_and(depths >= bounds[0], depths <= bounds[1]))
        return inds
        
    def get_nearest_tind(self, var, point, use_cache=True):
        assert var in self._current_variables
        time = self.gettimevar(var, use_cache)
        return time.nearest_index(point.time)
        
    def get_nearest_zind(self, var, point, use_cache=True):
        assert var in self._current_variables
        depths = self.getdepthvar(var, use_cache)
        return depths.nearest_index(point.depth)
        
    def __str__(self):
        k = []
        for key in self._current_variables:
            k.append(key)
        out = """
[[ 
  <Paegan Dataset Object>
  Dataset Type: """ + self._datasettype + """ 
  Resource: """ + self._filename + """
  Variables: 
  """ + str(k) + """
]]"""
        return out 
    
    def get_coord_names(self, var=None, **kwargs):
        assert var in self._current_variables
        ncvar = self.nc.variables[var]
        try:
            coordinates = ncvar.coordinates.split()
        except StandardError:
            coordinates = []
        # If the coordinate names not in kwargs, then figure
        # out the remaining coordinate names
        if "xname" in kwargs:
            xname = kwargs["xname"]
        else:
            xname = list(set(coordinates) & set(self._possiblex))
            if len(xname) > 0:
                xname = xname[0]
            else:
                xname = None
        if "yname" in kwargs:
            yname = kwargs["yname"]
        else:
            yname = list(set(coordinates) & set(self._possibley))
            if len(yname) > 0:
                yname = yname[0]
            else:
                yname = None
        if "zname" in kwargs:
            zname = kwargs["zname"]
        else:
            zname = list(set(coordinates) & set(self._possiblez))
            if len(zname) > 0:
                zname = zname[0]
            else:
                zname = None
        if "tname" in kwargs:
            tname = kwargs["tname"]
        else:
            tname = list(set(coordinates) & set(self._possiblet))
            if len(tname) > 0:
                tname = tname[0]
            else:
                tname = None      
        names = {"tname":tname, "zname":zname,
                 "xname":xname, "yname":yname}
        # find how the shapes match up to var
        # (should i use dim names or just sizes to figure out?)
        # I'm going to use dim names
        dims = ncvar.dimensions
        ndim = ncvar.ndim
        shape = ncvar.shape
        total = []
        for i in names:
            name = names[i]
            if name != None:
                cdims = self.nc.variables[name].dimensions
                for cdim in cdims:
                    try:
                        total.append(dims.index(cdim))
                    except StandardError:
                        pass 
        total = np.unique(np.asarray(total))

        for missing in range(ndim):
            if missing not in total:

                missing_dim = dims[missing]

                if missing_dim in self.nc.variables:
                    if missing_dim in self._possiblex:
                        name2 = "xname"
                    elif missing_dim in self._possibley:
                        name2 = "yname"
                    elif missing_dim in self._possiblez:
                        name2 = "zname"
                    elif missing_dim in self._possiblet:
                        name2 = "tname"
                    else:
                        name2 = None
                    if name2 != None:
                        names[name2] = missing_dim
        # Need to add next check if there are any dims left
        # to find variables with different names that use soley
        # those dims and appear in our type keys
        
        # then make guesses based on attributes like:
        # units, standard_name, perhaps time attribute, axis 
        # attribute (X,Y,Z,T),
        
        # then...
        return names
              
    def get_coord_dict(self, var=None, **kwargs):
        assert var in self._current_variables
        timevar = self.gettimevar(var)
        depthvar = self.getdepthvar(var)
        gridobj = self.getgridobj(var)
        return {"time":timevar, "z":depthvar, "xy":gridobj}
        
        
    def get_varname_from_stdname(self, standard_name=None,
        match=None):
        var_matches = []
        if match == None:
            for var in self._current_variables:
                try:
                    sn = self.nc.variables[var].standard_name
                    if standard_name == sn:
                        var_matches.append(var)
                except StandardError:
                    pass
        else:
            pass
        return var_matches
        
    def __repr__(self):
        s = "CommonDataset(" + self._filename + \
            ", dataset_type='" + self._datasettype + "')"
        return s
        
    def sub_coords(self, var, zbounds=None, bbox=None,
        timebounds=None, zinds=None, timeinds=None):
        assert var in self._current_variables
        coord_dict = self.get_coord_dict(var)
        names = self.get_coord_names(var)
        x, y, z, time = None, None, None, None
        if names['tname'] != None:
            #tname = names['tname']
            if timebounds != None:
                timeinds = self.get_tind_from_bounds(var, timebounds)
            time = coord_dict['time'][timeinds[0]:timeinds[-1]+1]
        if names['zname'] != None:
            #zname = names['zname']
            if zbounds != None:
                zinds = self.get_zind_from_bounds(var, zbounds)
            z = coord_dict['z'][zinds[0]:zinds[-1]+1]
        xinds, yinds = self.get_xyind_from_bbox(var, bbox)
        xy = coord_dict['xy']
        if names['xname'] != None:
            #xname = names['xname']
            if len(xy._xarray.shape) == 2:
                x = xy._xarray[xinds[0][0]:xinds[0][-1]+1, xinds[1][0]:xinds[1][-1]+1]
            elif len(xy.xarray.shape) == 1:
                x = xy._xarray[np.squeeze(xinds)]
        if names['yname'] != None:
            #yname = names['yname']
            if len(xy._yarray.shape) == 2:
                y = xy._yarray[yinds[0][0]:yinds[0][-1]+1, yinds[1][0]:yinds[1][-1]+1]
            elif len(xy.yarray.shape) == 1:
                y = xy._yarray[np.squeeze(yinds)]
        return subs(x=x, y=y, z=z, time=time)
        
    def get_indices(self, var, zbounds=None, bbox=None,
        timebounds=None, zinds=None, timeinds=None, 
        point=None, use_local=False, **kwargs):
        """
        
        Get smallest chunck of data that encompasses the 4-d
        bounding box limits of the data completely.
        
        
        """
        assert var in self._current_variables
        ncvar = self.nc.variables[var]
        names = self.get_coord_names(var)
        # find how the shapes match up to var
        # (should i use dim names or just sizes to figure out?)
        # I'm going to use dim names
        dims = ncvar.dimensions
        ndim = ncvar.ndim
        shape = ncvar.shape
        positions = dict()
        total = []
        for i in names:
            name = names[i]
            if i == "tname":
                common_name = "time"
            elif i == "zname":
                common_name = "z"
            elif i == "xname":
                common_name = "x"
            elif i == "yname":
                common_name = "y"
            else:
                common_name = None
            if common_name != None:
                positions[common_name] = None
                if name != None:
                    positions[common_name] = []
                    cdims = self.nc.variables[name].dimensions
                    for cdim in cdims:
                        try:
                            positions[common_name].append(dims.index(cdim))
                        except StandardError:
                            pass

        if positions["time"] != None:
            if timebounds != None:
                tinds = self.get_tind_from_bounds(var, timebounds)
            else:
                if timeinds == None:
                    if point != None:
                        tinds = np.asarray([self.get_nearest_tind(var, point)])
                    else:
                        tinds = np.arange(0, ncvar.shape[positions["time"]]+1)
                else:
                    tinds = timeinds
        if positions["z"] != None:
            if zbounds != None:
                zinds = self.get_zind_from_bounds(var, zbounds)
            else:
                if zinds == None:
                    if point != None:
                        zinds = np.asarray([self.get_nearest_zind(var, point)])
                    else:
                        zinds = np.arange(0, ncvar.shape[positions["z"]]+1)
                else:
                    pass
        if bbox != None:
            xinds, yinds = self.get_xyind_from_bbox(var, bbox)
        else:
            if point != None:
                num = kwargs.get("num", 1)
                xinds, yinds = self.get_xyind_from_point(var, point, num=num)
            else:
                xinds = [np.arange(0, ncvar.shape[pos]+1) for pos in positions["x"]]
                yinds = [np.arange(0, ncvar.shape[pos]+1) for pos in positions["y"]]

        indices = [None for i in range(ndim)]
        for name in positions:
            if positions[name] != None:
                if name == "time":
                    for i,position in enumerate(positions[name]):
                        indices[position] = tinds[i] 
                elif name == "z":
                    for i,position in enumerate(positions[name]):
                        indices[position] = zinds[i]
                elif name == "y":
                    for i,position in enumerate(positions[name]):
                        indices[position] = yinds[i]
                elif name == "x":
                    for i,position in enumerate(positions[name]):
                        indices[position] = xinds[i]
        
        return indices
    
    def get_values(self, var, zbounds=None, bbox=None,
        timebounds=None, zinds=None, timeinds=None, 
        point=None, use_local=False, **kwargs):
        """
        
        Get smallest chunck of data that encompasses the 4-d
        bounding box limits of the data completely.
        
        
        """
        assert var in self._current_variables
        ncvar = self.nc.variables[var]
        names = self.get_coord_names(var)
        # find how the shapes match up to var
        # (should i use dim names or just sizes to figure out?)
        # I'm going to use dim names
        dims = ncvar.dimensions
        ndim = ncvar.ndim
        shape = ncvar.shape
        positions = dict()
        total = []
        for i in names:
            name = names[i]
            if i == "tname":
                common_name = "time"
            elif i == "zname":
                common_name = "z"
            elif i == "xname":
                common_name = "x"
            elif i == "yname":
                common_name = "y"
            else:
                common_name = None
            if common_name != None:
                positions[common_name] = None
                if name != None:
                    positions[common_name] = []
                    cdims = self.nc.variables[name].dimensions
                    for cdim in cdims:
                        try:
                            positions[common_name].append(dims.index(cdim))
                        except StandardError:
                            pass
        # get t inds, z inds, xy inds
        # tinds = [[1,],]
        # zinds = [[1,],]
        # xinds = [[50,], [50,]]
        # yinds = [[50,], [50,]]
        if positions["time"] != None:
            if timebounds != None:
                tinds = self.get_tind_from_bounds(var, timebounds)
            else:
                if timeinds == None:
                    if point != None:
                        tinds = np.asarray([self.get_nearest_tind(var, point)])
                    else:
                        tinds = np.arange(0, ncvar.shape[positions["time"]]+1)
                else:
                    tinds = timeinds
        if positions["z"] != None:
            if zbounds != None:
                zinds = self.get_zind_from_bounds(var, zbounds)
            else:
                if zinds == None:
                    if point != None:
                        zinds = np.asarray([self.get_nearest_zind(var, point)])
                    else:
                        zinds = np.arange(0, ncvar.shape[positions["z"]]+1)
                else:
                    pass
        if bbox != None:
            xinds, yinds = self.get_xyind_from_bbox(var, bbox)
        else:
            if point != None:
                num = kwargs.get("num", 1)
                xinds, yinds = self.get_xyind_from_point(var, point, num=num)
            else:
                xinds = [np.arange(0, ncvar.shape[pos]+1) for pos in positions["x"]]
                yinds = [np.arange(0, ncvar.shape[pos]+1) for pos in positions["y"]]
        #if len(tinds) > 0 and len(zinds) > 0 and \
        #    len(xinds) > 0 and len(yinds) > 0:
        # Now take time inds, z inds, x and y inds and put them 
        # into the request in the right places:
       
        indices = [None for i in range(ndim)]
        for name in positions:
            if positions[name] != None:
                if name == "time":
                    for i,position in enumerate(positions[name]):
                        indices[position] = tinds[i] 
                elif name == "z":
                    for i,position in enumerate(positions[name]):
                        indices[position] = zinds[i]
                elif name == "y":
                    for i,position in enumerate(positions[name]):
                        indices[position] = yinds[i]
                elif name == "x":
                    for i,position in enumerate(positions[name]):
                        indices[position] = xinds[i]
        
        #logger.info("Getting data for %s with indexes: %s" % (var, str(indices)))

        if np.all([i.size >0 for i in indices]):
            data = self._get_data(var, indices, use_local)
        else:
            # data = None
            raise ValueError("no data inside the domian specified")
        return data
            
    def _get_data(self, var, **kwargs):
        raise NotImplementedError
                
    _get_values = get_values
    _getgridobj = getgridobj
    _gettimevar = gettimevar
    _getdepthvar = getdepthvar
    _lon2ind = lon2ind
    _ind2lon = ind2lon
    _lat2ind = lat2ind
    _ind2lat = ind2lat
    __get_data = _get_data
    
    """
    
        Methods that operate on and return another Dataset object.
        Intended to be strung together.
    
    """
    def restrict_bbox(self, bbox = None, **kwargs):
        raise NotImplementedError
            
    def restrict_time(self, times = None):
        assert times != None
        assert len(times) == 2
        new = self._copy()
        for var in new._current_variables:
            time_dimension = new.gettimevar(var)
            if time_dimension != None:
                inds = new.get_tind_from_bounds(var, times)
                time_dimension = _sub_by_nan(time_dimension, inds[0][0])
                new._coordcache[var].t = time_dimension
        return new
            
    def restrict_vars(self, varlist = None):
        assert varlist != None
        coord_names = []
        new = self._copy()
        if type(varlist) == str:
            varlist = (varlist,)
        for var in new._current_variables:
            coord_names = coord_names + new.get_coord_names(var).values()
        for var in self._current_variables:
            if (not var in set(varlist)) and (not var in set(coord_names)):
                new._current_variables.remove(var)
        return new
         
    def restrict_depth(self, depths = None):
        assert depths != None
        assert len(depths) == 2
        new = self._copy()
        for var in new._current_variables:
            depth_dimension = new.getdepthvar(var)
            if depth_dimension != None:
                inds = new.get_zind_from_bounds(var, depths)
                depth_dimension = _sub_by_nan(depth_dimension, inds[0])
                new._coordcache[var].z = depth_dimension
        return new
            
    def regrid(self, **kwargs):
        """
            TODO: Implement a generic regridding method.
                  Pass in lat/lon/depth/time arrays?
        """
        raise NotImplementedError
        
    def nearest_point(self, point):
        raise NotImplementedError
            
    def nearest_depth(self, depth):
        new = self._copy()
        if type(depth) != Location4D:
            depth = Location4D(depth=depth, latitude=0, longitude=0)
        for var in self._current_variables:
            depth_dimension = new.getdepthvar(var)
            if depth_dimension != None:
                ind = new.get_nearest_zind(var, depth)
                depth_dimension = _sub_by_nan(depth_dimension, ind)
                new._coordcache[var].z = depth_dimension 
        return new
            
    def nearest_time(self, time):
        new = self._copy()
        if type(time) != Location4D:
            time = Location4D(time=time, latitude=0, longitude=0)
        for var in self._current_variables:
            time_dimension = new.gettimevar(var)
            if time_dimension != None:
                ind = new.get_nearest_tind(var, time)
                time_dimension = _sub_by_nan(time_dimension, ind)
                new._coordcache[var].t = time_dimension
        return new    



import numpy as np
import netCDF4
from paegan.utils.asagreatcircle import AsaGreatCircle
from paegan.location4d import Location4D
from shapely.geometry import MultiLineString, LineString
from shapely.ops import polygonize

class Gridobj:
    def __init__(self, nc, xname=None, yname=None,
        xunits=None, yunits=None, projected=False, **kwargs):
        self._projected = projected
        if type(nc) is str:
            nc = netCDF4.Dataset(nc)
        
        self._nc = nc
        self._xname = xname
        self._yname = yname
        self._ndim = self._nc.variables[self._xname].ndim
        self._ymesh = None
        self._xmesh = None
        self._type = None
        
        if self._xname != None:
            self._x_nc = self._nc.variables[self._xname]
            self._xarray = np.asarray(self._x_nc[:])
            if self.xmax <= 360 and self.xmin >= 0:
                self._xarray[np.where(self._xarray > 180)] = \
                    self._xarray[np.where(self._xarray > 180)] - 360
            
        else:
            self._xarray = np.asarray((),)
        if self._yname !=None:
            self._y_nc = self._nc.variables[self._yname]
            self._yarray = np.asarray(self._y_nc[:])
        else:
            self._yarray = np.asarray((),)

    
    def get_xbool_from_bbox(self, bbox):
        return np.logical_and(self._xarray<=bbox[2],                
                              self._xarray>=bbox[0])
        
    def get_ybool_from_bbox(self, bbox):
        return np.logical_and(self._yarray<=bbox[3],                
                              self._yarray>=bbox[1])
        
    def getydata(self):
        pass
        
    def getxdata(self):
        pass
        
    def findx(self):
        pass
        
    def findy(self):
        pass
    
    def get_xmax(self):
        return np.nanmax(np.nanmax(self._xarray))
        
    def get_ymax(self):
        return np.nanmax(np.nanmax(self._yarray))
        
    def get_xmin(self):
        return np.nanmin(np.nanmin(self._xarray))
        
    def get_ymin(self):
        return np.nanmin(np.nanmin(self._yarray))
    
    def get_bbox(self):
        """
            TODO: Implement ncell bbox
        """
        if self._ndim == 2:
            xtmp = self._xarray[np.isnan(self._xarray)==False]
            bbox = np.nanmin(self._xarray[:,0]), self.ymin, np.nanmax(self._xarray[:,-1]), self.ymax
        else:
            xtmp = self._xarray[np.isnan(self._xarray)==False]
            bbox = xtmp[0], self.ymin, xtmp[-1], self.ymax
        return bbox

    def get_boundingpolygon(self):
        """
            TODO: Implement ncell bbox

            order of lines creation (assumes 0,0 is x)
            -----3-----
            |         |
            4         2
            |         |
            x----1-----
        """

        if self._ndim == 2: # CGRID
            nx,ny = self._xarray.shape
            one = MultiLineString([((self._xarray[i][0],self._yarray[i][0]),(self._xarray[i+1][0],self._yarray[i+1][0])) for i in range(nx-1)])
            two = MultiLineString([((self._xarray[nx-1][j],self._yarray[nx-1][j]),(self._xarray[nx-1][j+1],self._yarray[nx-1][j+1])) for j in range(ny-1)])
            three = MultiLineString([((self._xarray[i][ny-1],self._yarray[i][ny-1]),(self._xarray[i-1][ny-1],self._yarray[i-1][ny-1])) for i in reversed(range(1,nx))])
            four = MultiLineString([((self._xarray[0][j],self._yarray[0][j]),(self._xarray[0][j-1],self._yarray[0][j-1])) for j in reversed(range(1,ny))])
            m = one.union(two).union(three).union(four)
        else: # RGRID
            nx,ny = self._xarray.shape[0], self._yarray.shape[0]
            one = LineString([(self._xarray[i], self._yarray[0]) for i in range(nx)])
            two = LineString([(self._xarray[-1], self._yarray[i]) for i in range(ny)])
            three = LineString([(self._xarray[i], self._yarray[-1]) for i in reversed(range(nx))])
            four = LineString([(self._xarray[0], self._yarray[i]) for i in reversed(range(ny))])
            m = MultiLineString([one,two,three,four])

        polygons = list(polygonize(m))
        # -- polygonize returns a list of polygons, including interior features, the largest in area "should" be the full feature
        polygon = sorted(polygons, key=lambda x: x.area)[-1]
        return polygon
            
    def get_projectedbool(self):
        return self._projected
    
    def get_xunits(self):
        try:
            units = self._nc.variables[self._xname].units
        except StandardError:
            units = None
        return units
        
    def get_yunits(self):
        try:
            units = self._nc.variables[self._yname].units
        except StandardError:
            units = None
        return units
        
    def bbox_to_wkt(self):
        pass
    
    def near_xy(self, **kwargs):
        """
            TODO: Implement ncell near_xy
        """
        point = kwargs.get("point", None)
        if point == None:
            lat = kwargs.get("lat", None)
            lon = kwargs.get("lon", None)
            point = Location4D(latitude=lat, longitude=lon)
        num = kwargs.get("num", 1)
        ncell = kwargs.get("ncell", False)
        if ncell:
            if num > 1:
                pass
            else:
                distance = AsaGreatCircle.great_distance(
                    start_lats=self._yarray, start_lons=self._xarray,
                    end_lats=point.latitude, end_lons=point.longitude)["distance"]
                inds = np.where(distance == np.nanmin(distance))
                xinds, yinds = inds, inds
        else:
            if self._ndim == 2:
                distance = AsaGreatCircle.great_distance(
                    start_lats=self._yarray, start_lons=self._xarray,
                    end_lats=point.latitude, end_lons=point.longitude)["distance"]
                yinds, xinds = np.where(distance == np.nanmin(distance))
            else:
                #if self._xmesh == None and self._ymesh == None:
                #    self._xmesh, self._ymesh = np.meshgrid(self._xarray, self._yarray)
                if num > 1:
                    minlat = np.abs(self._yarray - point.latitude)
                    minlon = np.abs(self._xarray - point.longitude)
                    lat_cutoff = np.sort(minlat)[num-1]
                    lon_cutoff = np.sort(minlon)[num-1]
                elif num == 1:
                    lat_cutoff = np.nanmin(np.abs(self._yarray - point.latitude))
                    lon_cutoff = np.nanmin(np.abs(self._xarray - point.longitude))
                yinds = np.where(np.abs(self._yarray - point.latitude) <= lat_cutoff)
                xinds = np.where(np.abs(self._xarray - point.longitude) <= lon_cutoff)

        return yinds, xinds 

        
    is_projected = property(get_projectedbool, None)
    xmax = property(get_xmax, None)
    ymax = property(get_ymax, None)
    xmin = property(get_xmin, None)
    ymin = property(get_ymin, None)
    bbox = property(get_bbox, None)
    boundingpolygon = property(get_boundingpolygon, None)
    xunits = property(get_xunits, None)
    yunits = property(get_yunits, None)
    _findy = findy
    _findx = findx
    _getxdata = getxdata
    _getydata = getydata
    
    



class Feature(object):

    def __init__(self):
        pass

    def get_type(self):
        return self._type
    def set_type(self, type):
        self._type = value
    type = property(get_type, set_type)

    def get_title(self):
        return self._location
    def set_title(self, title):
        self._title = title
    title = property(get_title, set_title)

    def get_description(self):
        return self._description
    def set_description(self, description):
        self._description = description
    description = property(get_description, set_description)



import math

class GreatCircle(object):
    # -----------------------------------------------------------------------
    # | Algrothims from Geocentric Datum of Australia Technical Manual      |
    # |                                                                     |
    # | http://www.anzlic.org.au/icsm/gdatum/chapter4.html                  |
    # |                                                                     |
    # | This page last updated 11 May 1999                                  |
    # |                                                                     |
    # | Computations on the Ellipsoid                                       |
    # |                                                                     |
    # | There are a number of formulae that are available                   |
    # | to calculate accurate geodetic positions,                           |
    # | azimuths and distances on the ellipsoid.                            |
    # |                                                                     |
    # | Vincenty's formulae (Vincenty, 1975) may be used                    |
    # | for lines ranging from a few cm to nearly 20,000 km,                |
    # | with millimetre accuracy.                                           |
    # | The formulae have been extensively tested                           |
    # | for the Australian region, by comparison with results               |
    # | from other formulae (Rainsford, 1955 & Sodano, 1965).               |
    # |                                                                     |
    # | * Inverse problem: azimuth and distance from known                  |
    # |                     latitudes and longitudes                        |
    # | * Direct problem: Latitude and longitude from known                 |
    # |                     position, azimuth and distance.                 |
    # | * Sample data                                                       |
    # | * Excel spreadsheet                                                 |
    # |                                                                     |
    # | Vincenty's Inverse formulae                                         |
    # | Given: latitude and longitude of two points                         |
    # |                     (phi1, lembda1 and phi2, lembda2),              |
    # | Calculate: the ellipsoidal distance (s) and                         |
    # | forward and reverse azimuths between the points (alpha12, alpha21). |
    # |                                                                     |
    # -----------------------------------------------------------------------

    @staticmethod
    def vinc_dist(  f,  a,  phi1,  lembda1,  phi2,  lembda2 ) :
        """ 

        Returns the distance between two geographic points on the ellipsoid
        and the forward and reverse azimuths between these points.
        lats, longs and azimuths are in radians, distance in metres 

        Returns ( s, alpha12,  alpha21 ) as a tuple

        """

        if (abs( phi2 - phi1 ) < 1e-8) and ( abs( lembda2 - lembda1) < 1e-8 ) :
          return 0.0, 0.0, 0.0
  
        two_pi = 2.0*math.pi

        b = a * (1.0 - f)

        TanU1 = (1-f) * math.tan( phi1 )
        TanU2 = (1-f) * math.tan( phi2 )
        
        U1 = math.atan(TanU1)
        U2 = math.atan(TanU2)

        lembda = lembda2 - lembda1
        last_lembda = -4000000.0                # an impossibe value
        omega = lembda

        # Iterate the following equations, 
        #  until there is no significant change in lembda 
        
        while ( last_lembda < -3000000.0 or lembda != 0 and abs( (last_lembda - lembda)/lembda) > 1.0e-9 ) :
        
          sqr_sin_sigma = pow( math.cos(U2) * math.sin(lembda), 2) + \
                pow( (math.cos(U1) * math.sin(U2) - \
                math.sin(U1) *  math.cos(U2) * math.cos(lembda) ), 2 )

          Sin_sigma = math.sqrt( sqr_sin_sigma )
          
          Cos_sigma = math.sin(U1) * math.sin(U2) + math.cos(U1) * math.cos(U2) * math.cos(lembda)
          
          sigma = math.atan2( Sin_sigma, Cos_sigma )

          Sin_alpha = math.cos(U1) * math.cos(U2) * math.sin(lembda) / math.sin(sigma)
          alpha = math.asin( Sin_alpha )
          
          Cos2sigma_m = math.cos(sigma) - (2 * math.sin(U1) * math.sin(U2) / pow(math.cos(alpha), 2) )
          
          C = (f/16) * pow(math.cos(alpha), 2) * (4 + f * (4 - 3 * pow(math.cos(alpha), 2)))
          
          last_lembda = lembda
          
          lembda = omega + (1-C) * f * math.sin(alpha) * (sigma + C * math.sin(sigma) * \
                (Cos2sigma_m + C * math.cos(sigma) * (-1 + 2 * pow(Cos2sigma_m, 2) )))
        

        u2 = pow(math.cos(alpha),2) * (a*a-b*b) / (b*b)
        
        A = 1 + (u2/16384) * (4096 + u2 * (-768 + u2 * (320 - 175 * u2)))
        
        B = (u2/1024) * (256 + u2 * (-128+ u2 * (74 - 47 * u2)))
        
        delta_sigma = B * Sin_sigma * (Cos2sigma_m + (B/4) * \
                (Cos_sigma * (-1 + 2 * pow(Cos2sigma_m, 2) ) - \
                (B/6) * Cos2sigma_m * (-3 + 4 * sqr_sin_sigma) * \
                (-3 + 4 * pow(Cos2sigma_m,2 ) )))
        
        s = b * A * (sigma - delta_sigma)
        
        alpha12 = math.atan2( (math.cos(U2) * math.sin(lembda)), \
                (math.cos(U1) * math.sin(U2) - math.sin(U1) * math.cos(U2) * math.cos(lembda)))
        
        alpha21 = math.atan2( (math.cos(U1) * math.sin(lembda)), \
                (-math.sin(U1) * math.cos(U2) + math.cos(U1) * math.sin(U2) * math.cos(lembda)))

        if ( alpha12 < 0.0 ) : 
                alpha12 =  alpha12 + two_pi
        if ( alpha12 > two_pi ) : 
                alpha12 = alpha12 - two_pi

        alpha21 = alpha21 + two_pi / 2.0
        if ( alpha21 < 0.0 ) : 
                alpha21 = alpha21 + two_pi
        if ( alpha21 > two_pi ) : 
                alpha21 = alpha21 - two_pi

        return s, alpha12,  alpha21 
        
        
    #----------------------------------------------------------------------------
    # Vincenty's Direct formulae                                                |
    # Given: latitude and longitude of a point (phi1, lembda1) and              |
    # the geodetic azimuth (alpha12)                                            |
    # and ellipsoidal distance in metres (s) to a second point,                 |
    #                                                                           |
    # Calculate: the latitude and longitude of the second point (phi2, lembda2) |
    # and the reverse azimuth (alpha21).                                        |
    #                                                                           |
    #----------------------------------------------------------------------------
    @staticmethod
    def vinc_pt( f, a, phi1, lembda1, alpha12, s ) :
        """

        Returns: lat and long of projected point and reverse azimuth,
        given a reference point and a distance and azimuth to project.
        lats, longs and azimuths are passed in RADIANS

        Returns ( phi2,  lambda2,  alpha21 ) as a tuple, all in radians

        """

        two_pi = 2.0*math.pi

        if ( alpha12 < 0.0 ) : 
                alpha12 = alpha12 + two_pi
        if ( alpha12 > two_pi ) : 
                alpha12 = alpha12 - two_pi

        
        b = a * (1.0 - f)

        TanU1 = (1-f) * math.tan(phi1)
        U1 = math.atan( TanU1 )
        sigma1 = math.atan2( TanU1, math.cos(alpha12) )
        Sinalpha = math.cos(U1) * math.sin(alpha12)
        cosalpha_sq = 1.0 - Sinalpha * Sinalpha
        
        u2 = cosalpha_sq * (a * a - b * b ) / (b * b)
        A = 1.0 + (u2 / 16384) * (4096 + u2 * (-768 + u2 * \
                (320 - 175 * u2) ) )
        B = (u2 / 1024) * (256 + u2 * (-128 + u2 * (74 - 47 * u2) ) )
        
        # Starting with the approximation
        sigma = (s / (b * A))

        # Not moving anywhere. We can return the location that was passed in.
        if sigma == 0:
            return phi1, lembda1, alpha12

        last_sigma = 2.0 * sigma + 2.0  # something impossible
        
        # Iterate the following three equations 
        # until there is no significant change in sigma 

        # two_sigma_m , delta_sigma

        while ( abs( (last_sigma - sigma) / sigma) > 1.0e-9 ) :

           two_sigma_m = 2 * sigma1 + sigma
           
           delta_sigma = B * math.sin(sigma) * ( math.cos(two_sigma_m) \
                        + (B/4) * (math.cos(sigma) * \
                        (-1 + 2 * math.pow( math.cos(two_sigma_m), 2 ) -  \
                        (B/6) * math.cos(two_sigma_m) * \
                        (-3 + 4 * math.pow(math.sin(sigma), 2 )) *  \
                        (-3 + 4 * math.pow( math.cos (two_sigma_m), 2 ))))) \
           
           last_sigma = sigma
           sigma = (s / (b * A)) + delta_sigma
        
        
        phi2 = math.atan2 ( (math.sin(U1) * math.cos(sigma) + math.cos(U1) * math.sin(sigma) * math.cos(alpha12) ), \
                ((1-f) * math.sqrt( math.pow(Sinalpha, 2) +  \
                pow(math.sin(U1) * math.sin(sigma) - math.cos(U1) * math.cos(sigma) * math.cos(alpha12), 2))))
        

        lembda = math.atan2( (math.sin(sigma) * math.sin(alpha12 )), (math.cos(U1) * math.cos(sigma) -  \
                math.sin(U1) *  math.sin(sigma) * math.cos(alpha12)))
        
        C = (f/16) * cosalpha_sq * (4 + f * (4 - 3 * cosalpha_sq ))
        
        omega = lembda - (1-C) * f * Sinalpha *  \
                (sigma + C * math.sin(sigma) * (math.cos(two_sigma_m) + \
                C * math.cos(sigma) * (-1 + 2 * math.pow(math.cos(two_sigma_m),2) )))
        
        lembda2 = lembda1 + omega
        
        alpha21 = math.atan2 ( Sinalpha, (-math.sin(U1) * math.sin(sigma) +  \
                math.cos(U1) * math.cos(sigma) * math.cos(alpha12)))

        alpha21 = alpha21 + two_pi / 2.0
        if ( alpha21 < 0.0 ) :
                alpha21 = alpha21 + two_pi
        if ( alpha21 > two_pi ) :
                alpha21 = alpha21 - two_pi


        return phi2,  lembda2,  alpha21 


import itertools
import collections

from paegan.cdm.dsg.collections.base.feature_collection import FeatureCollection
from paegan.cdm.dsg.collections.base.point_collection import PointCollection
from paegan.cdm.dsg.features.base.point import Point
from paegan.utils.asalist import AsaList

class NestedPointCollection(FeatureCollection):
    """
        A collection of PointCollections
    """

    def __init__(self, **kwargs):
        super(NestedPointCollection,self).__init__(**kwargs)

    def calculate_bounds(self):
        """
            Calculate the time_range, bbox, and size of this collection.
            Will scan all data.
            Ensures that .size, .bbox and .time_range return non-null.

            If the collection already knows its bbox; time_range; and/or size,
            they are recomputed.
        """ 
        single_point_collection = PointCollection(elements=list(AsaList.flatten(self)))
        single_point_collection.calculate_bounds()
        self.bbox = single_point_collection.bbox
        self.time_range = single_point_collection.time_range
        self.depth_range = single_point_collection.depth_range
        self._point_size = single_point_collection.size
        self.size = len(self._elements)
        

    def flatten(self):
        """
            Returns a Generator of Points that are part of this collection
        """
        return AsaList.flatten(self)

    def get_point_size(self):
        """
            Returns the number of actual Points in this NestedPointCollection

            Ex.  pc = 10 profiles with 20 bins each will return 200
                 pc.size = 10
                 pc.point_size = 200
        """
        return self._point_size
    point_size = property(get_point_size, None)


import copy

import numpy as np

from paegan.cdm.dataset import Dataset, _sub_by_nan


class NCellDataset(Dataset):
    """
    
        NCellDataset(Dataset)
    
    """
    def __init__(self, *args,**kwargs):
        super(NCellDataset,self).__init__(*args, **kwargs)
        if None in self.nc.variables:
            self._is_topology = True
            self.topology_var_name = None
    
    def _copy(self):
        new = NCellDataset(self._filename, self._datasettype)
        new._coordcache = copy.copy(self._coordcache)
        new._current_variables = copy.copy(self._current_variables)
        return new
        
    def restrict_bbox(self, bbox = None, **kwargs):
        assert bbox != None
        assert len(bbox) == 4
        new = self._copy()
        for var in new._current_variables:
            grid = new.getgridobj(var)
            if grid != None:
                inds, inds =  new.get_xyind_from_bbox(var, bbox)
                grid._xarray = _sub_by_nan(grid._xarray, inds[0][0])
                grid._yarray = _sub_by_nan(grid._yarray, inds[0][0])
                new._coordcache[var].xy = grid
        return new
    
    def nearest_point(self, point):
        assert type(point) == Location4D
        new = self._copy()
        for var in new._current_variables:
            grid = new.getgridobj(var)
            if grid != None:
                inds, inds = new.get_xyind_from_point(var, point)
                grid._xarray = _sub_by_nan(grid._xarray, inds)
                grid._yarray = _sub_by_nan(grid._yarray, inds)
                new._coordcache[var].xy = grid
        return new
        
    def get_xyind_from_bbox(self, var, bbox):
        grid = self.getgridobj(var)
        xbool = grid.get_xbool_from_bbox(bbox)
        ybool = grid.get_ybool_from_bbox(bbox)
        inds = np.where(np.logical_and(xbool, ybool))
        return inds, inds #xinds, yinds
        
    def get_xyind_from_point(self, var, point, **kwargs):
        num = kwargs.get("num", 1)
        grid = self.getgridobj(var)
        inds, inds = grid.near_xy(point=point, num=num, ncell=True)
        return inds, inds

    def _get_data(self, var, indarray, use_local=False):
        ndims = len(indarray)
        if use_local == False:
            var =    self.nc.variables[var]
        else:
            pass
        if ndims == 1:
            data = var[:]
            data = data[indarray]
        elif ndims == 2:
            data = var[:]
            data = data[indarray[0], indarray[1]]
        elif ndims == 3:
            data = var[indarray[0], indarray[1], :]
            data = data[:, :, indarray[2]]
        elif ndims == 4:
            data = var[indarray[0], indarray[1], indarray[2], 
                       indarray[3]]
        elif ndims == 5:
            data = var[indarray[0], indarray[1], indarray[2], 
                       indarray[3], indarray[4]]
        elif ndims == 6:
            data = var[indarray[0], indarray[1], indarray[2], 
                       indarray[3], indarray[4], indarray[5]]
        return data



from paegan.cdm.dsg.collections.base.point_collection import PointCollection

class Profile(PointCollection):
    """
        A collection of points along a vertical (z) access
        ie. A single profile from an ADCP or CTD

        A profile has a nominal lat/lon and time.
        Actual time may be constant, or vary with z.
        The z coordinates are monotonic, and may be increasing or decreasing.
    """
    def __init__(self, **kwargs):
        super(Profile,self).__init__(**kwargs)
        self._type = "Profile"

    def get_location(self):
        return self._location
    def set_location(self, location):
        self._location = location
    location = property(get_location, set_location)

    def get_time(self):
        """
            Nominal time of the profile.  This is usually the first time a reading was taken,
            since a profile's time can be constant or vary with z.
        """
        return self._time
    def set_time(self, time):
        self._time = time
    time = property(get_time, set_time)


from paegan.cdm.dsg.collections.base.nested_point_collection import NestedPointCollection

class SectionCollection(NestedPointCollection):
    """
        A collection of Sections
    """

    def __init__(self, **kwargs):
        super(SectionCollection,self).__init__(**kwargs)



from paegan.cdm.dsg.features.base.feature import Feature

class Point(Feature):
    """
        A collection of observations at one time and location.
        ie. An instantaneous reading of multiple parameters from a sensor
    """

    def __init__(self):
        self._member_names = []
        self._members = []
        self._type = "Point"
        self._time = None

    def get_location(self):
        """
            A Shapely Point object (x,y,z) in the cartesian plane
        """
        return self._location
    def set_location(self, location):
        self._location = location
    location = property(get_location, set_location)

    def get_time(self):
        """
            The observation time
        """
        return self._time
    def set_time(self, time):
        self._time = time
    time = property(get_time, set_time)

    def add_member(self, member):
        self._members.append(member)
        self._member_names.append(member['name'])

    def get_member(self, **kwargs):
        """
            Get Member by variable name
        """
        return self._members[self._member_names.index(kwargs.get('name'))]

    def get_member_names(self):
        return self._member_names
    def set_members(self, members):
        self._members = members
    def get_members(self):
        return self._members
    members = property(get_members, set_members)
